program WinDir;
{ ----------------------------------------------------------------------
  Description : Display a full screen directory listing, showing
                File name, Extension, Size, Date and Time
  Compiler    : Turbo Pascal for Windows
  System      : MS-DOS 3.31
  Notice      : Copyright (c) 1991 by Pacific Software
  Date        : 1991 June 12
  -------------------------------------------------------------------- }
uses
  WObjects, WinTypes, WinProcs, WinDos, Strings, StdDlgs;

{$R WINDIR.RES}

const
  MaxDirSize     = 512;
  fieldLen       = 20;
  MonthStr       : array[1..12, 0..3] of Char = (
   'Jan', 'Feb', 'Mar', 'Apr', 'May', 'Jun',
   'Jul', 'Aug', 'Sep', 'Oct', 'Nov', 'Dec');
  cm_Path        = 101; { Menu Identifiers }
  cm_Quit        = 102;
  cm_About       = 103;
  id_path        = 101; { Dialog box fields}
  id_NameButton  = 102;
  id_ExtButton   = 103;
  id_SizeButton  = 104;
  id_TimeButton  = 105;
  id_ChkHidden   = 106;
  id_ChkModified = 107;

type
  PDirEntry      = ^TDirEntry;
  TDirEntry      = record
                     Attr           : Byte;
                     Time           : LongInt;
                     Size           : LongInt;
                     Name           : array[0..12] of Char;
                     Ext            : array[0..3] of Char;
                   end;
  TDirList       = array[0..MaxDirSize - 1] of PDirEntry;
  LessFunc       = function(X, Y : PDirEntry) : Boolean;

  { Dialog transfer record }
  ItemTransferBuffer = record
                         PathStr        : array[0..fieldLen - 1] of Char;
                         NameButton     : Bool;
                         ExtButton      : Bool;
                         SizeButton     : Bool;
                         TimeButton     : Bool;
                         ChkHidden      : Bool;
                         ChkModified    : Bool;
                       end;

  { The dialog is used for input of new path parameters. }
  PPathDialog   = ^TPathDialog;
  TPathDialog   = object(TDialog)
                   PathEdit      : PEdit;
                   Radio         : PRadioButton;
                   Check         : PCheckBox;
                   constructor Init(AParent : PWindowsObject; ATitle : PChar);
                 end;

  { Declare TScrollWindow, a TWindow descendant }
  PScrollWindow  = ^TScrollWindow;
  TScrollWindow =
    object(TWindow)
      NotBacked      : Boolean;
      SearchAttr     : Byte;
      Count          : Integer;
      Path           : array[0..fsPathName] of Char;
      DirList        : TDirList;
      Less           : LessFunc;           { sort order }
      ItemBuffer     : ItemTransferBuffer; { for PathDialog }
      constructor Init(ATitle : PChar);
      procedure GetWindowClass(var WndClass : TWndClass); virtual;
      procedure GetCommand;
      procedure FindFiles;
      procedure QuickSort(L, R : Integer);

      { Menu responses }
      procedure cmPath(var Msg : TMessage); virtual cm_First + cm_Path;
      procedure cmClose(var Msg : TMessage); virtual cm_First + cm_Quit;
      procedure cmAbout(var Msg : TMessage); virtual cm_First + cm_About;
        
      { Windows Message Responses }
      procedure Paint(PaintDC : HDC; var PaintInfo : TPaintStruct); virtual;
      procedure wmKeyDown(var Msg : TMessage); virtual wm_KeyDown;
    end;

  { Declare TScrollApp, a TApplication descendant }
  TScrollApp = object(TApplication)
                 procedure InitMainWindow; virtual;
               end;

  { --------------------------------------------------
    Utility procedures / functions
    ------------------------------------------------- }

  function LessName(X, Y : PDirEntry) : Boolean; FAR;
  { sort the directory by name }
  begin
    LessName := X^.Name < Y^.Name;
  end; { LessName }


  function LessExt(X, Y : PDirEntry) : Boolean; FAR;
  { sort the directory by extension }
  begin
    LessExt := X^.Ext < Y^.Ext;
  end; { LessExt }


  function LessSize(X, Y : PDirEntry) : Boolean; FAR;
  { sort the directory by file size }
  begin
    LessSize := X^.Size < Y^.Size;
  end; { LessSize }


  function LessTime(X, Y : PDirEntry) : Boolean; FAR;
  { sort the directory by time and date }
  begin
    LessTime := X^.Time < Y^.Time;
  end; { LessTime }


  { --------------------------------------------------
    TPathDialog method implementations:             
    ------------------------------------------------ }

  { The edit and radio controls will contain the transfer data. }
  constructor TPathDialog.Init(AParent : PWindowsObject; ATitle : PChar);
  begin
    TDialog.Init(AParent, ATitle);
    New(PathEdit, initResource(@Self, id_path, fieldLen));
    New(Radio, initResource(@Self, id_NameButton));
    New(Radio, initResource(@Self, id_ExtButton));
    New(Radio, initResource(@Self, id_SizeButton));
    New(Radio, initResource(@Self, id_TimeButton));
    New(Check, initResource(@Self, id_ChkHidden));
    New(Check, initResource(@Self, id_ChkModified));
  end;


  { --------------------------------------------------
    TScrollWindow method implementations:
    ------------------------------------------------- }

  constructor TScrollWindow.Init(ATitle : PChar);
  { TScrollWindow constructor, sets scroll styles and constructs
    the Scroller object. }
  begin
    SearchAttr := faReadOnly + faDirectory + faArchive;
    GetCommand;
    FindFiles;
    TWindow.Init(nil, 'WinDir');
    Attr.Style := WS_Caption or WS_SysMenu or WS_MinimizeBox or ws_VScroll;
    Scroller := New(PScroller, Init(@Self, 8, 15, 80, 100));
    StrCopy(ItemBuffer.PathStr, Path);
    ItemBuffer.NameButton := True;
    ItemBuffer.ExtButton := False;
    ItemBuffer.SizeButton := False;
    ItemBuffer.TimeButton := False;
    ItemBuffer.ChkHidden := False;
    ItemBuffer.ChkModified := False; 
  end; { Init }


  procedure TScrollWindow.GetWindowClass(var WndClass : TWndClass);
  { assign style, define menu, load icon }
  begin
    TWindow.GetWindowClass(WndClass);
    WndClass.Style := 0; 
    WndClass.lpszMenuName := 'Menu'; 
    WndClass.hIcon := LoadIcon(hInstance, 'Sailin');
  end; { GetWindowClass }

  procedure TScrollWindow.GetCommand;
  { interpret the command line. }
  var
    I, J           : Integer;
    s              : String;

  begin { GetCommand }
    NotBacked := False;
    Less := LessName; { set up default sort order }
    Path[0] := #0;
    for I := 1 to ParamCount do begin
      s := ParamStr(I);
      if s[1] in ['/', '-'] then
        case Upcase(s[2]) of
          'N' : Less := LessName;
          'E' : Less := LessExt;
          'S' : Less := LessSize;
          'T' : Less := LessTime;
          'H' : SearchAttr := faReadOnly+faHidden+faSysFile+faDirectory;
          'M' : NotBacked := True;
        else
          WriteLn('Invalid option: ', s[J]);
        end
      else
        StrPCopy(Path, s);
    end; { for }

    { special case when drive only is specified }
    if (Length(s) = 2) and (s[2] = ':') then begin
      GetDir(Ord(Upcase(s[1])) - 64, s);
      StrPCopy(Path, s);
    end;
  end; { GetCommand }


  procedure TScrollWindow.FindFiles;
  { scan the directory for name, ext, size, date, time }
  label
    Loop;

  var
    N              : Word;
    Dir            : array[0..fsDirectory] of Char;
    Name           : array[0..fsFileName] of Char;
    Ext            : array[0..fsExtension] of Char;
    F              : file;
    Attribute      : Word;
    SearchRec      : TSearchRec;

  begin { FindFiles }
    FileExpand(Path, Path);
    if Path[strLen(Path) - 1] <> '\' then begin
      Assign(F, Path);
      GetFAttr(F, Attribute);
      if (DosError = 0) and (Attribute and faDirectory <> 0) then
        StrLCat(Path, '\', fsPathName);
    end;
    FileSplit(Path, Dir, Name, Ext);
    if Name[0] = #0 then StrCopy(Name, '*');
    if Ext[0] = #0 then StrCopy(Ext, '.*');
    StrECopy(StrECopy(StrECopy(Path, Dir), Name), Ext);

    Count := 0;
    FindFirst(Path, SearchAttr, SearchRec);
    while (DosError = 0) and (Count < MaxDirSize) do begin
      IF Notbacked THEN IF (SearchRec.attr AND faArchive) = 0 THEN GOTO Loop;
      IF SearchRec.name[0] = '.' THEN GOTO Loop; { ignore . and .. entries }
      N := strLen(SearchRec.Name) + 10;
      GetMem(DirList[Count], N);
      Move(SearchRec.Attr, DirList[Count]^, N);
      Inc(Count);
Loop :
      FindNext(SearchRec);
    end;
    if Count <> 0 then QuickSort(0, Count - 1);
  end; { FindFiles }


  procedure TScrollWindow.QuickSort(L, R : Integer);
  { sort the directory by name, ext, size or date }
  var
    I, J           : Integer;
    X, Y           : PDirEntry;

  begin { QuickSort }
    I := L;
    J := R;
    X := DirList[(L + R) div 2];
    repeat
      while Less(DirList[I], X) do Inc(I);
      while Less(X, DirList[J]) do Dec(J);
      if I <= J then begin
        Y := DirList[I];
        DirList[I] := DirList[J];
        DirList[J] := Y;
        Inc(I);
        Dec(J);
      end;
    until I > J;
    if L < J then QuickSort(L, J);
    if I < R then QuickSort(I, R);
  end; { QuickSort }


  procedure TScrollWindow.cmPath(var Msg : TMessage);
  { Get a new path, Find files, repaint the window. }
  var
    Dlg            : TPathDialog;
    Reply          : Integer;

  begin { cmPath }
    Dlg.Init(@Self, 'PathDlg');
    Dlg.TransferBuffer := @ItemBuffer;
    Reply := Dlg.Execute;
    Dlg.Done;
    if Reply = id_Ok then begin { change directory and find the new files }
      StrCopy(Path, ItemBuffer.PathStr);
      if ItemBuffer.NameButton then Less := LessName;
      if ItemBuffer.ExtButton then Less := LessExt;
      if ItemBuffer.SizeButton then Less := LessSize;
      if ItemBuffer.TimeButton then Less := LessTime;
      if ItemBuffer.ChkHidden then SearchAttr := $0017
      else  SearchAttr := faReadOnly + faDirectory + faArchive;
      NotBacked := ItemBuffer.ChkModified;
(***) IF NotBacked THEN MessageBox (HWindow, 'NotBacked = True', 'Status', mb_OK);
      FindFiles;
      InvalidateRect(HWindow, nil, True);
    end;
  end; { cmPath }


  procedure TScrollWindow.cmClose(var Msg : TMessage);
  { End the Program. }
  begin
    postQuitMessage(0);
  end;


  procedure TScrollWindow.cmAbout(var Msg : TMessage);
  { Bring up the About box. }
  begin
    Application^.ExecDialog(New(PDialog, Init(@Self, 'About')));
  end;


  procedure TScrollWindow.Paint(PaintDC : HDC; var PaintInfo : TPaintStruct);
  { paint the window with the file information }
  var
    Rowcount, jj, HalfWay : Integer;
    XPos, YPos            : Integer;
    TotalSize, BytesFree  : LongInt;
    P                     : array[0..2] of PChar;
    Q                     : PChar;
    T                     : TDateTime;
    N                     : array[0..fsFileName] of Char;
    E                     : array[0..fsExtension] of Char;
    si                    : array[0..7] of Char;
    s, S1, S2, S3, s4     : array[0..79] of Char;

  begin { Paint }
    SelectObject(PaintDC, getStockObject(ansi_fixed_font));
    ShowWindow(HWindow, sw_Maximize);  
    XPos := 10;
    YPos := 0;
    HalfWay := Count DIV 2;
    StrCopy(s, 'Directory of ');
    StrCat(s, Path);
    SetWindowText(HWindow, s); { re-write the caption }
    if Count = 0 then begin
      MessageBox(HWindow, 'No matching files', 'Result', mb_OK);
      Exit;
    end;
    TotalSize := 0;
    Rowcount := 0;
    for jj := 0 to Count - 1 do with DirList[jj]^ do begin
      Q := StrPos(Name, '.');
      if (Q = nil) or (Q = Name) then begin
        StrCopy(N, Name);
        StrCopy(E, '');
      end
      else begin
        StrLCopy(N, Name, Q - Name);
        StrCopy(E, Q + 1);
      end;
      P[0] := N;
      P[1] := E;
      if Attr and faDirectory <> 0 then si := '<DIR>   '
      else Str(Size:8, si);
      P[2] := si;
      WVSPrintF(S1, '%-9.9s%-3.3s%8.8s', P);
      UnpackTime(Time, T);
      T.Year := T.Year mod 100;
      WVSPrintF(S2, ' %02d', T.Year);
      Q := MonthStr[T.Month];
      WVSPrintF(S3, '-%3s', Q);
      WVSPrintF(s4, '-%02d %02d:%02d', T.Day);
      s := S1;
      StrCat(s, S2);
      StrCat(s, S3);
      StrCat(s, s4);
      TextOut(PaintDC, XPos, YPos, s, strLen(s));
      Inc(TotalSize, Size);
      Inc(YPos, 15); { advance to the next line }
      Inc(Rowcount);
      if Rowcount > HalfWay then begin
        if XPos = 10 then XPos := 310 else XPos := 10;
        YPos := 0;
        Rowcount := 0;
      end;
    end; { for }
    WVSPrintF(s, ' %10d Files', Count);
    TextOut(PaintDC, XPos, YPos, s, strLen(s));
    Inc (YPos, 15);
    WVSPrintF(s, ' %10ld Total size', TotalSize);
    TextOut(PaintDC, XPos, YPos, s, strLen(s));
    BytesFree := DiskFree(Ord(Path[0]) - 64);
    Inc (YPos, 15);
    WVSPrintF(s, ' %10ld Bytes free', BytesFree);
    TextOut(PaintDC, XPos, YPos, s, strLen(s));
  end; { Paint }


  procedure TScrollWindow.wmKeyDown(var Msg : TMessage);
  { change the default key down behaviour for selected keys }
  var
    key            : Word;
  begin
    key := Msg.WParam;
    case key of
      VK_UP : Scroller^.ScrollBy(0, - 1);
      VK_DOWN : Scroller^.ScrollBy(0, 1);
      VK_Prior : Scroller^.ScrollBy(0, - 25);
      VK_Next : Scroller^.ScrollBy(0, 25);
      VK_HOME : Scroller^.ScrollTo(0, 0);
      VK_END : Scroller^.ScrollTo(0, 100);
      VK_Escape : postQuitMessage(0);
    end; { case }
  end; { KeyDown }


  { --------------------------------------------------
    TScrollApp method implementations:             
    ------------------------------------------------- }

  procedure TScrollApp.InitMainWindow;
  { Construct the TScrollApp's MainWindow of type TScrollWindow }
  begin
    MainWindow := New(PScrollWindow, Init('Directory'));
  end;


  { --------------------------------------------------
    Main Program
    ------------------------------------------------- }
var
  ScrollApp      : TScrollApp;

begin { WinDir }
  ScrollApp.Init('WinDir');
  ScrollApp.Run;
  ScrollApp.Done;
end. { WinDir }

